/*
        GPKMAKE.C

        by Marcus Grber 1996
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gpk.h"

char filebuf[DEFAULT_BLOCK_SIZE];

char *getline_low(char *buf,int size,FILE *f)
{
        int i;

        if(fgets(buf,size,f))
        {
          for(i=strlen(buf)-1;
              i>=0 && (buf[i]=='\n' || buf[i]==' ' || buf[i]=='\t');
              i--)
            ;                           /* remove trailing whitespaces */
          buf[i+1]=0;                   /* terminate string */
          for(i=0; buf[i]==' ' || buf[i]=='\t'; i++)
            ;                           /* skip leading whitespaces */
          return buf+i;                 /* return string (without whites) */
        }
        return NULL;                    /* error getting line */
}

char *getline(char *buf,int size,FILE *f)
{
        char *p;

        do {
          p=getline_low(buf,size,f);    /* get one line from file */
        } while(p && (*p=='\0' || *p=='#'));
                                        /* skip over blank lines or comments */
        return p;                       /* return line or NULL if eof */
}

long writestr(FILE *f,char *str)
{
        long pos=ftell(f);              /* starting position of string */
        char eos=0;

        if(str)
          fwrite(str,strlen(str)+1,1,f);
        else
          fwrite(&eos,sizeof(eos),1,f);
        return pos;
}

int main(int argc,char *argv[])
{
        FILE *fScript=NULL, *fGPK=NULL, *fData;
        char buf[512],*line,*p,*q;
        struct GPK_header hdr={ID_VALUE,x1_VALUE,device_VALUE,x2_VALUE};
        long *pos_file=NULL,size,n_blk,blksize;
        unsigned n;

        puts("\nGPKMake 1.02 -=- by Marcus Grber 1996/97\n");
        if(argc<2)
        {
          puts("Usage: GPKMAKE <script_file>");
          return 255;
        }

        fScript = fopen(argv[1],"r");
        if(!fScript)
        {
          puts("Error opening script file.");
          return 1;
        }

        /*** Get line with GPK filename */
        if( !(line=getline(buf,sizeof(buf),fScript)) ) goto error;
        fGPK = fopen(line,"wb");
        if(!fScript)
        {
          puts("Error creating GPK file.");
          return 2;
        }

        /*** Get line specifying the number of files */
        if( !(line=getline(buf,sizeof(buf),fScript)) ) goto error;
        hdr.n_files = atoi(line);
        pos_file = malloc(sizeof(long)*hdr.n_files);
        fseek(fGPK,sizeof(hdr)+sizeof(long)*hdr.n_files,SEEK_SET);
                                        /* leave space for header */

        /*** Get line with file title */
        if( !(line=getline(buf,sizeof(buf),fScript)) ) goto error;
        hdr.pos_title = writestr(fGPK,line);

        /*** Get line with file description */
        if( !(line=getline(buf,sizeof(buf),fScript)) ) goto error;
        hdr.pos_desc = writestr(fGPK,line);

        /*** Get lines with file names, pack files */
        n=0;                            /* no files read yet */
        while( (line=getline(buf,sizeof(buf),fScript)) && n<hdr.n_files )
        {
          p=strtok(line," \t");         /* parse filename */
          printf("Adding %s...",line);

          if(*p!='*')                   /* not an empty file? */
            fData=fopen(p,"rb");        /* attempt to open source file */
          else
            fData=NULL;

          p=strtok(NULL," \t");         /* parse destination pathname */
          if(fData)
          {
            q=strrchr(p,'\\');          /* find last backslash */
            if(q==NULL) q=p; else q++;  /* no backslash: use entire string */
          }
          else                          /* file not found: empty file */
            q = "";                     /* write empty string */

          pos_file[n] = writestr(fGPK,q);
                                        /* write filename, store starting pos */
          writestr(fGPK,p);             /* write pathname */

          p=strtok(NULL," \t");         /* parse version string */
          writestr(fGPK,p);             /* write version string */

          if(fData)                     /* file was successfully opened */
          {
            fseek(fData,0,SEEK_END);    /* get size of data file */
            size=ftell(fData);
            fseek(fData,0,SEEK_SET);
            if(size)
            {
              fwrite(&size,sizeof(size),1,fGPK);
                                        /* write size in bytes */
              n_blk = ((size+DEFAULT_BLOCK_SIZE-1)/DEFAULT_BLOCK_SIZE);
                                        /* number of blocks holding file */
              fwrite(&n_blk,sizeof(n_blk),1,fGPK);
                                        /* write number of blocks */
            }
            while(size)                 /* still data left? */
            {
              blksize=min(DEFAULT_BLOCK_SIZE,size);
                                        /* size of following block */
              fwrite(&blksize,sizeof(blksize),1,fGPK);
                                        /* prefix block with size */
              fread(filebuf,1,(int)blksize,fData);
              fwrite(filebuf,1,(int)blksize,fGPK);
                                        /* copy one buffer's worth of data */
              size-=blksize;            /* count down remaining size */
            }
            fclose(fData);
            putchar('\n');
          }
          else                          /* file not present: use placeholder */
          {
            size=0;                     /* write out zero size */
            fwrite(&size,sizeof(size),1,fGPK);
            printf(" not found - creating empty file.\n");
          }

          n++;                          /* count files */
        }

        if(n!=hdr.n_files || *line)
        {
          puts("Number of files specified in script is not correct.");
          hdr.n_files=n;
        }

        fseek(fGPK,0,SEEK_SET);         /* back to start of file */
        fwrite(&hdr,sizeof(hdr),1,fGPK);/* write header */
        fwrite(pos_file,sizeof(long),n,fGPK);

error:  if(fGPK) fclose(fGPK);
        if(fScript) fclose(fScript);

        if(pos_file)                    /* release memory for file pointers */
          free(pos_file);
}
